var util = require('util');
var events = require('events');
var path = require('path');
var fs = require('fs');
var os = require('os');

var _ = require('lodash');
var async = require('async');

var global = require('./global');
var childManager = require('./child-manager');

var ProcessWatcher = require('./watch-process');
var checker = require('./checker');

var YeomanRunner = require('./yeoman-runner');
var GruntRunner = require('./grunt-runner');


/**
 * ABC操作类，一个应用会生成一个ABC实例
 *
 * @fires ABC#stdout 子进程控制台输出
 * @fires ABC#error 子进程错误信息
 * @fires ABC#child-exit 子进程退出
 * @fires ABC#event 收到事件消息
 * @fires ABC#yeoman:resolvable yeoman消息
 * @fires ABC#message 消息
 *
 * @param config
 * @constructor
 */
function ABC (config) {

    events.EventEmitter.apply(this, arguments);

    this.root = config.root;
    this.type = config.type || null;

    process.nextTick(function(){
        this.check();
    }.bind(this));

    this.once('check end', function(error) {

        if (error) {
            this.updateStatus('check-with-error');
            return;
        }

        this.updateStatus('ready');
        this.emit('ready');
        this.getYeomanRunner();

    }.bind(this));
}

ABC._gu_msg_id = 0;
ABC._gu_msg_cb_list = {};

util.inherits(ABC, events.EventEmitter);

/**
 * check files and npm modules
 */
ABC.prototype.check = function check() {

    this.once('npm end', function(error) {
        this.emit('check end');

        if (error) {
            this.emit('error', error);
        }
    });

    this.updateStatus('checking');

    this.checkFile();

    // check files
    if (!this.packageJSON ||
        !this.abcJSON ||
        !this.abcJSON.type) {

        this.updateStatus('not-a-abc-project');
        this.updateStatus('ready');

        this.emit('ready');
        return;
    }

    this.checkNPM(function(err) {

        if (err) {
            this.emit('error', {
                message: 'NPM install error',
                origin: err
            });

            this.emit('npm end', err);

            return;
        }

        this.emit('npm end');

    }.bind(this));
};

/**
 * checking and load abc.json package.json
 */
ABC.prototype.checkFile = function checkFile() {

    var files = {};

    var abcJSONFile = path.join(this.root, 'abc.json');
    var packageJSONFile = path.join(this.root, 'package.json');

    files.abcJSON = fs.existsSync(abcJSONFile);
    files.packageJSON = fs.existsSync(packageJSONFile);

    if (files.abcJSON) {
        try {
            this.abcJSON = JSON.parse(fs.readFileSync(abcJSONFile));

        } catch(e) {
            this.emit('error', {
                message: 'JSON parse error:' + abcJSONFile,
                origin: e
            });
            return;
        }
    }

    if (files.packageJSON) {
        try {
            this.packageJSON = JSON.parse(fs.readFileSync(packageJSONFile));
        } catch(e) {
            this.emit('error', {
                message: 'JSON parse error:' + packageJSONFile,
                origin: e
            });
        }
    }

};

ABC.prototype.checkGenerator = function checkGenerator(callback) {
    this.updateStatus('check-generator');
    if (!this.type) {
        callback(null);
        return;
    }

    var gm = require('./generator-manager');

    if (gm.exists(this.type)) {
        this.updateStatus('check-generator-exist');
        callback(null);
        return;
    }
    gm.install(this.type, callback);

};


/**
 * check and install NPM package
 * @param callback {Function}
 */
ABC.prototype.checkNPM = function checkNPM(callback) {

    var self = this;

    self.updateStatus('checking-npm');
    var child = checker.npm(this.root, callback);
    if (child && child.pid) {
        self.emit('process:new', {
            title: '安装项目 NPM 依赖',
            child: child
        });
    }

};


ABC.prototype.runGrunt = function runGrunt(tasks, options, callback) {
    var g = new GruntRunner({
        tasks: tasks,
        options: options
    });
    g.run(callback);
};

ABC.prototype.exitGrunt = function exitGrunt(child_id) {
    childManager.kill(child_id);
};



ABC.prototype.getType = function getType() {
    return this.abcJSON ? this.abcJSON.type : this.type || '';
};

ABC.prototype.getName = function getName() {
    return this.abcJSON ? this.abcJSON.name : '';
};

ABC.prototype.info = function info() {
    return {
        root: this.root,
        name: this.getName(),
        type: this.getType(),
        pkg: this.packageJSON,
        abc: this.abcJSON
    }
};

/**
 * 杀死子进程
 */
ABC.prototype.destroy = function destroy() {

    this.emit('destroy');
    this._destroyed = true;

    if (this.yeomanRunner) {
        this.yeomanRunner.destroy();
    }

    if (this.gruntRunner) {
        this.gruntRunner.destroy();
    }

    this.removeAllListeners();

};

ABC.prototype.getYeomanRunner = function getYeomanRunner() {
    if (!this.yeomanRunner || this.yeomanRunner.destroyed) {

        this.yeomanRunner = new YeomanRunner({
            root: this.root,
            type: this.getType()
        });

        this.yeomanRunner.on('child:ready', function() {
            this.yeomanRunner.scan();
        }.bind(this));

        this.emit('new-yeoman', this.yeomanRunner);
    }

    return this.yeomanRunner;
};

ABC.prototype.getGruntRunner = function getGruntRunner() {

    if (!this.gruntRunner || this.yeomanRunner.destroyed) {
        this.gruntRunner = new GruntRunner({
            root: this.root
        });
        this.emit('new-grunt', this.gruntRunner);
    }

    return this.gruntRunner;
};


ABC.prototype.watchChild = function watchChild(child) {
    ProcessWatcher.watch(child, this);
};

/**
 * update status
 * @param status {String}
 */
ABC.prototype.updateStatus = function updateStatus(status) {

    if (!status) {
        return;
    }
    this.status = status;

    this.emit('status:update', {
        status: status
    });
};


module.exports = ABC;

