/**
 * @overview 用Socket处理与客户端的通信
 */

var ABC = require( '../lib/abc');
var _ = require('lodash');
var global = require('./global');
var fs = require('fs');
var mkdirp = require('mkdirp');
var file = require('./utils/file');

var system = require('./utils/system');

var ABCHome = system.getUserHome('.abc');

var GeneratorManager = require('./generator-manager');

var pkg = require('../package.json');

var history = require('./history');


exports.handle = function (io) {

    // Client 和 Server建立连接
    io.sockets.on('connection', function (socket) {

        var session = {
            childPool: {}
        };

        //连接断开，删除子进程
        socket.on('disconnect', function() {

            killMgr();

            if (session.grunt) {
                session.grunt.destroy();
                session.grunt = null;
            }

            if (session.yeoman) {
                session.yeoman.destroy();
                session.yeoman = null;
            }

            // kill the child process
            _.keys(session.childPool).forEach(function (pid) {

                if (pid) {
                    session.childPool[pid].kill();
                }

            });

        });


        //helper
        function getMgr() {
            return session.mgr;
        }

        function newMgr(config) {
            var prev = getMgr();

            if (prev) {
                killMgr();
            }

            var mgr = new ABC(config);

            session.mgr = mgr;

            mgr
                .on('new-grunt', function (runner) {
                    watchGrunt(runner);
                })

                .on('new-yeoman', function (runner) {
                    watchYeoman(runner);
                })

                .on('event', function (data) {
                    socket.emit('action:' + data.name, data.attrs);
                })
                .on('error', function(error) {
                    var errorObj = {};

                    if (error instanceof Error) {
                        errorObj = error.message;
                        errorObj.origin = error;
                    } else {
                        errorObj = error
                    }

                    if (!errorObj.message) {
                        errorObj.message = 'Unknown Error'
                    }

                    socket.emit('action:error', errorObj);
                })
                .on('status:update', function (data) {
                    socket.emit('action:status:update', data);
                })
                .on('process:new', function (data) {
                    addChild(data);
                })
                .once('ready', function () {
                    var info = mgr.info();

                    history.add({
                        root: info.root,
                        name: info.name,
                        type: info.type
                    });

                    socket.emit('ready');
                });

            return mgr;
        }

        function killMgr() {
            var mgr = getMgr();

            if (mgr) {
                mgr.destroy();
                session.mgr = null;
            }
        }

        function watchYeoman(runner) {
            if (session.yeoman && session.yeoman !== runner) {
                session.yeoman.destroy();
            }

            session.yeoman = runner;
            watchRunner(runner, 'yeoman:', ['ready', 'start', 'end', 'error', 'resolvable', 'scan']);
        }

        function watchGrunt(runner) {
            if (session.grunt && session.grunt === runner) {
                return;
            }

            if (session.grunt) {
                session.grunt.destroy();
                session.grunt = null;
            }

            session.grunt = runner;

            watchRunner(runner, 'grunt:', ['start', 'end', 'error']);
        }

        function watchRunner(runner, prefix, events) {

            events.forEach(function(event) {
                var socketEvent = prefix + event;
                runner.on('child:' + event, function(data) {

                    if (typeof data === 'undefined') {
                        socket.emit(socketEvent);
                    } else {
                        socket.emit(socketEvent, data);
                    }
                });
            });

            runner.on('stdout', function(str){
                socket.emit(prefix + 'stdout', str);
            });
        }

        function addChild(option) {
            var child, title;

            if (option.pid) {
                child = option;
                title = 'Untitled'
            } else {
                child = option.child;
                title = option.title;
            }



            var pid = child.pid;

            socket.emit('process:new', {
                title: title,
                pid: pid
            });

            session.childPool[pid] = child;

            child.on('exit', function(){
                socket.emit('process:exit:'+pid);
                delete session.childPool[pid];
            });

            if(child.stdout) {
                child.stdout.setEncoding();
                child.stdout.on('data', function (data) {
                    socket.emit('process:stdout:' + pid, data);
                });
            }

            if(child.stderr) {
                child.stderr.setEncoding();
                child.stderr.on('data', function (data) {
                    socket.emit('process:stdout:'+pid, data);
                });
            }
        }

        socket.on('appInfo', function() {

            socket.emit('appInfo', {
                pkg: pkg
            });
        });



        //切换工作目录
        socket.on('cd', function (data, respond) {
            var root = data.root;

            if (!root) {
                respond({
                    status: 'error',
                    message: 'No root path given.'
                });
                return;
            }

            if (!fs.existsSync(root)) {

                respond({
                    status: 'error',
                    message: 'Director is not exist'
                });

                return;
            }

            killMgr();

            try {

                newMgr({
                    root: root
                });

                respond({
                    'status': 'ok'
                });

            } catch (e) {

                respond({
                    'status': 'error',
                    'message': 'Can not switch to: ' + root,
                    'error' : e.message
                });

            }
        });

        socket.on('update:core', function (callback) {
            var spawn = require('./utils/spawn-command');

            var child = spawn('npm', ['update', 'abc-web-core'], {
                cwd: ABCHome
            }, function (err) {
                if (err) {
                    callback(err);
                    console.error(err);
                    return;
                }
                // restart process
                process.exit(9);
                callback();
            });

            addChild({
                title: "更新主程序",
                child: child
            });



        });

        socket.on('init', function (data) {


            var root = data.root;
            var type = data.type;

            if (!file.isPathAbsolute(root)) {

                socket.emit('error', {
                    message: "需要输入目录的绝对路径",
                    title: "目录初始化失败"
                });
                return;

            }

            if (!fs.existsSync(root)) {
                try {
                    mkdirp.sync(root);
                } catch(e) {
                    socket.emit('error', {
                        message: "无法创建目录，权限不足或其他原因",
                        title: "目录初始化失败"
                    });
                    return;
                }
            }

            // clear npm check cache
            var checker = require('./checker');
            checker.clearCache(root);

            var YeomanRunner = require('./yeoman-runner');

            var yr = new YeomanRunner({
                root: root,
                type: type
            });

            watchYeoman(yr);

            yr.on('child:ready', function() {
                yr.run();
            });

        });

        [   'yeoman:run',
            'yeoman:exit',
            'yeoman:resolve',
            'yeoman:scan',
            'yeoman:ui'
        ].
            forEach(function(evName){
                socket.on(evName, function(data, callback) {
                    if (!session.yeoman) {
                        socket.emit('error', {
                            message: 'No yeoman for this session'
                        });
                        return;
                    }

                    var method = evName.split(':')[1];

                    if (!session.yeoman) {
                        socket.emit('error', {
                            message: 'session has no yeoman'
                        });
                        return;
                    }

                    if (!session.yeoman[method]) {
                        socket.emit('error', {
                            message: 'yeoman has no method ' + method
                        });
                        return;
                    }

                    session.yeoman[method].apply(session.yeoman, arguments);

                });
            });

        socket.on('grunt:run', function (data) {

            var mgr = getMgr();

            if (!mgr) {

                socket.emit('error', {
                    title: "Grunt",
                    message: 'Not in project'
                });

                return;
            }

            var grunt = mgr.getGruntRunner();

            grunt.run(data.tasks, data.options);
        });

        // Exit Grunt
        socket.on('grunt:exit', function(pid){
            process.kill(pid);
        });

        //获取历史记录
        socket.on('history', function(){
            socket.emit('history', history.get());
        });

        //获取历史记录
        socket.on('history:rm', function(root){
            history.remove(root);
            socket.emit('history', history.get());
        });

        // 获取当前应用信息
        socket.on('info', function (respond) {
            var mgr = getMgr();

            if (mgr.status !== 'ready') {
                return respond(null)
            }

            respond(mgr.info());

        });

        // get all generator
        socket.on('generators', function (respond) {
            require('./generators').getAll(respond);
        });

        socket.on('generators:list', function () {
            GeneratorManager.list(function (err, generators) {
                if (err) {
                    socket.emit('error', {
                        message: 'List generator failed'
                    });
                }
                socket.emit('generator:list', generators);
            });
        });

        socket.on('generator:install', function (name, respond) {
            var child = GeneratorManager.install(name, respond);
            addChild({
                title: "安装目录规范",
                child: child
            });
        });

        socket.on('generator:update', function (name, respond) {
            GeneratorManager.update(name, respond);
        });

        socket.on('open', function(something) {
            require('./utils/system').commandOpen(something, function(err) {
                if (err) {
                    socket.emit('error', {
                        message: something,
                        title: '无法打开目标'
                    });
                }
            });
        });

        socket.on('exit', function (respond) {

            if (!getMgr()) {
                respond();
                return;
            }

            killMgr();

            respond();

        });

        socket.on('restart', function() {
            process.exit(0);
        });

    });
};