var path = require('path');

var currentGenerator;
var generatorCache = {};

var abcRoot = process.env.ABC_ROOT;

var sendMessage = process.send ? function  (data) {
    process.send(data);
} : function() {};

sendMessage({
    msgType: 'connect'
});

function restartProcess(){
    process.exit(0);
}

/**
 * 具体执行u
 *
 * 通过 Fork的消息机制与 ABCMgr通信， 消息接口定义如下
 *
 * message 接口
 * message.type, String, grunt or yeoman 消息类型
 *
 * message.action: yeoman:init
 * message.attrs
 *   generator: kissy-pie:app
 *   params: Array
 */

process.on('message', messageHandler);


/**
 * Emit an Event
 * @param name {String}
 * @param attrs {Object}
 * @optional
 */
function sendEvent(name, attrs) {
    var msg = {
        msgType: 'event',
        raw: {
            name: name
        }

    };

    if (attrs) {
        msg.raw.attrs = attrs
    }

    sendMessage(msg);
}

/**
 * handle message from parent process
 * @param data {Object} message obj
 * @param data.action {String} action to exec
 * @param data.attr {Object} attributes for action
 */
function messageHandler(data) {

    var msgId = data._msg_id;

    var raw = data.raw;

    function respond (data) {
        sendMessage({
            type: 'respond',
            _msg_id: msgId,
            raw: data
        });
    }


    var action = raw.action;
    var attrs = raw.attrs;

    if (!action ) {
        return;
    }

    var actions = action.indexOf(':') === -1 ? [action] : action.split(':');

    var command = actions[0];
    var act = actions[1];


    if (!command) {
        return;
    }

    var handler;

    if (!act) {
        handler = handlers[command];
    } else if (handlers[command]) {
        handler =  handlers[command][act];
    }

    if (handler) {
        if (handler.length > 1) {
            handler(attrs, respond);
        } else {
            handler(attrs);
            respond();
        }

    } else {
        notSupported();
    }
}

function notSupported(data, done) {
    var message;

    if (!data.type) {
        message = 'no type provide!';
    } else {
        message = data.type + ' is not supported';
    }

    done({
        error: message,
        data: data
    });
}

function getGenerator (sub, type, params, options) {

    type = type || process.env['ABC_TYPE'];

    if (!type) {
        sendEvent('error', {
            message: "child: No type for generator!"
        });
        return null;
    }

    sub = sub || 'app';

    var generatorName = type + ':' + sub;

    if (generatorCache[generatorName]) {
        return generatorCache[generatorName];
    }

    var resolved = path.join(abcRoot, 'node_modules/generator-' + type + '/' + sub + '/index.js');

    var Generator = require(resolved);

    var generator = new Generator(params || [], {
        enableUI: true,
        resolved: resolved,
        env: {
            cwd: process.cwd()
        },
        name: generatorName
    });

    generatorCache[generatorName] = generator;

    return generator;

}

function scanApp(callback) {

    var generator = getGenerator('app');

    if (!generator) {
        sendEvent('error', {
            message: 'Can not find generator app'
        });
        return;
    }

    if (!generator.scan) {
        sendEvent('error', {
            message: 'Not  an UI enabled generator'
        });
        return;
    }

    generator.scan(function(app){
        sendEvent('scan', app);
        if (callback) callback();
    });
}

var handlers = {
    yeoman: {

        ui: function(attrs, callback) {

            var generator = getGenerator('app');
            callback(generator.getUI());

        },

        scan: function() {
            scanApp();
        },


        run: function(attrs) {

            var type = attrs.type;
            var subtype = attrs.sub || 'app';
            var args = attrs.args || [];
            var options = attrs.options;

            var generator = getGenerator(subtype, type, args, options);
            currentGenerator = generator;

            generator.on('start', function(){
                sendEvent('start');
            });

            generator.on('end', function(){
                scanApp(function() {
                    restartProcess();
                });
            });

            generator.on('conflict', function(conf) {
                sendEvent('resolvable', {
                    type: 'conflict',
                    conf: conf
                });
            });

            generator.on('prompt', function(prompts) {
                sendEvent('resolvable', {
                    type: 'prompts',
                    data: prompts
                });
            });

            generator.run();
        },

        exit: function () {
            sendEvent('exit');
            process.nextTick(function() {
                restartProcess();
            });

        },

        'resolve': function(attrs) {
            if (!currentGenerator) {
                sendEvent('error', {
                    message: 'Nothing can be resolved, generator is not running'
                });
                return;
            }

            if (attrs.type === 'prompt') {
                currentGenerator.resolvePrompt(null, attrs.props);
            }

            if (attrs.type === 'conflict') {
                currentGenerator.resolveConflict(attrs.filepath, attrs.method);
            }

        }
    },

    ping: function (attrs){
        sendEvent('pingback', attrs);
    },

    'kill': function () {
        process.exit();
    }
}
