var util = require('util');
var _ = require('lodash');
var system = require('./utils/system');
var generatorManager = require('./generator-manager');
var path = require('path');
var fork = require('child_process').fork;
var Runner = require('./runner');
var global = require('./global');

var abcChildPath = path.join(__dirname, 'yeoman-child.js');

module.exports = Yeoman;

function Yeoman(options) {
    Runner.apply(this);

    this.type = options.type;
    this.root = options.root;
    this._running = false;

    this._gu_msg_cb_list = {};

    this.once('check-done', function(data) {
        if (data && data.error) {
            this.emit('error', data.error);
            return;
        }
        this._initChild();
    });

    this.once('child-connect', function() {
        this.emit('child:ready');
    });

    this.on('child:start', function () {
        this._running = true;
    }.bind(this));

    this.on('child:end', function () {
        this._running = false;
    }.bind(this));

    process.nextTick(this._check.bind(this));
}

util.inherits(Yeoman, Runner);

Yeoman.prototype._check = function _check() {

    if (generatorManager.exists(this.type)) {
        this.emit('check-done');
        return;
    }

    generatorManager.install(this.type, function(err) {
        if (err) {
            this.emit('check-done', {
                error: err
            });
            return;
        }

        this.emit('check-done');

    }.bind(this));
};


Yeoman.prototype._initChild = function _initChild() {
    var childEnv = _.clone(process.env);

    childEnv.ABC_TYPE = this.type;
    childEnv.ABC_ROOT = system.getUserHome('.abc');

    this.child = fork(abcChildPath, {
        cwd: this.root,
        env: childEnv,
        stdio: '',
        silent: true
    });

    this.watchChild(this.child);

    this.child
        .on('message', this._messageHandler.bind(this))
        .on('error', function(error) {

            this.emit('child:error', {
                message: 'Could not start abc child',
                origin: error
            });

        }.bind(this))

        .on('exit', function () {
            this.emit('yeoman-child exit');
            this.child.removeAllListeners();

            if (this._running) {
                this.emit('child:end');
            }

            if (!this._destroyed) {
                this.emit('yeoman-child restart');
                this._initChild();
            }
        }.bind(this));

};

/**
 * 负责处理子进程传来的消息
 * @param data
 * @param data.msgType {String} 消息类型
 * @param data._msg_id {String} 消息id
 * @private
 */
Yeoman.prototype._messageHandler = function _messageHandler (data) {

    this.emit('message', data);

    if (data.msgType === 'connect') {
        this.emit('child-connect');
        return;
    }

    if (data.msgType === 'event') {
        var raw = data.raw || {};
        this.emit('child:' + raw.name, raw.attrs);
        return;
    }


    // 根据消息id，执行对应的回调
    if ('_msg_id' in data) {
        var msgId = data._msg_id;
        var msgList = this._gu_msg_cb_list;

        if (msgId in msgList) {
            var msgCb =  msgList[msgId];

            delete msgList[msgId];
            msgCb(data.raw);
        }
    }
};

Yeoman.prototype.ui = function ui(callback) {
    this._send({
        action: 'yeoman:ui'
    }, callback);
};

Yeoman.prototype.resolve = function resolve(data) {
    this._send({
        action: 'yeoman:resolve',
        attrs: data
    });
};

Yeoman.prototype.scan = function scan() {
    this._send({
        action: 'yeoman:scan'
    });
};

Yeoman.prototype.run = function scan(data) {

    data = data || {};

    if (_.isString(data.args)) {
        data.args = data.args.split(' ');
    }

    if (!data.sub || data.sub === 'app') {
        var npmCache = global.get('npmCache');
        npmCache = _.without(npmCache, this.root);
        global.set('npmCache', npmCache);
    }

    this._send({
        action: 'yeoman:run',
        attrs: data
    });
};

Yeoman.prototype.exit = function exit(callback) {
    this._send({
        action: 'yeoman:exit'
    });
};

Yeoman.prototype.newMsgId = function newMsgId () {
    if (!this._msg_id) {
        this._msg_id = 0;
    }
    this._msg_id ++;
    return this._msg_id;
};
/**
 * send message to abc child process
 * @param data {Object}
 * @param callback {Function}
 */
Yeoman.prototype._send = function _send(data, callback) {

    if (!this.child || !this.child.connected) {

        this.emit('error', {
            message: 'child not connected'
        });

        return;
    }

    var msgId = this.newMsgId();

    if (_.isFunction(callback)) {
        this._gu_msg_cb_list[msgId] = callback;
    }

    this.child.send({
        _msg_id: msgId,
        raw: data
    });
};

Yeoman.prototype.destroy = function destroy() {
    if (this._destroyed) {
        return;
    }
    this._destroyed = true;

    if(this.child && this.child.connected) {
        this.child.kill();
    }

    this.removeAllListeners();
};
